# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Nokia Networks.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from osprofiler import profiler
import traceback as tb

from mistral.db import utils as db_utils
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.engine import action_queue
from mistral.engine import tasks
from mistral.engine import workflow_handler as wf_handler
from mistral import exceptions as exc
from mistral.lang import parser as spec_parser
from mistral.services import scheduler
from mistral.workflow import base as wf_base
from mistral.workflow import commands as wf_cmds
from mistral.workflow import states


"""Responsible for running tasks and handling results."""

LOG = logging.getLogger(__name__)

_REFRESH_TASK_STATE_PATH = (
    'mistral.engine.task_handler._refresh_task_state'
)

_SCHEDULED_ON_ACTION_COMPLETE_PATH = (
    'mistral.engine.task_handler._scheduled_on_action_complete'
)

_SCHEDULED_ON_ACTION_UPDATE_PATH = (
    'mistral.engine.task_handler._scheduled_on_action_update'
)


@profiler.trace('task-handler-run-task', hide_args=True)
def run_task(wf_cmd):
    """Runs workflow task.

    :param wf_cmd: Workflow command.
    """

    task = _build_task_from_command(wf_cmd)

    try:
        task.run()
    except exc.MistralException as e:
        wf_ex = wf_cmd.wf_ex
        task_spec = wf_cmd.task_spec

        msg = (
            "Failed to run task [error=%s, wf=%s, task=%s]:\n%s" %
            (e, wf_ex.name, task_spec.get_name(), tb.format_exc())
        )

        LOG.error(msg)

        task.set_state(states.ERROR, msg)

        wf_handler.force_fail_workflow(wf_ex, msg)

        return

    if task.is_waiting() and (task.is_created() or task.is_state_changed()):
        _schedule_refresh_task_state(task.task_ex, 1)


@profiler.trace('task-handler-on-action-complete', hide_args=True)
def _on_action_complete(action_ex):
    """Handles action completion event.

    :param action_ex: Action execution.
    """

    task_ex = action_ex.task_execution

    if not task_ex:
        return

    task_spec = spec_parser.get_task_spec(task_ex.spec)

    wf_ex = task_ex.workflow_execution

    task = _create_task(
        wf_ex,
        spec_parser.get_workflow_spec_by_execution_id(wf_ex.id),
        task_spec,
        task_ex.in_context,
        task_ex
    )

    try:
        task.on_action_complete(action_ex)
    except exc.MistralException as e:
        wf_ex = task_ex.workflow_execution

        msg = ("Failed to handle action completion [error=%s, wf=%s, task=%s,"
               " action=%s]:\n%s" %
               (e, wf_ex.name, task_ex.name, action_ex.name, tb.format_exc()))

        LOG.error(msg)

        task.set_state(states.ERROR, msg)

        wf_handler.force_fail_workflow(wf_ex, msg)


@profiler.trace('task-handler-on-action-update', hide_args=True)
def _on_action_update(action_ex):
    """Handles action update event.

    :param action_ex: Action execution.
    """

    task_ex = action_ex.task_execution

    if not task_ex:
        return

    task_spec = spec_parser.get_task_spec(task_ex.spec)

    wf_ex = task_ex.workflow_execution

    task = _create_task(
        wf_ex,
        spec_parser.get_workflow_spec_by_execution_id(wf_ex.id),
        task_spec,
        task_ex.in_context,
        task_ex
    )

    try:
        task.on_action_update(action_ex)

        if states.is_paused(action_ex.state):
            wf_handler.pause_workflow(wf_ex)

        if states.is_running(action_ex.state):
            # If any subworkflow of the parent workflow is paused,
            # then keep the parent workflow execution paused.
            for task_ex in wf_ex.task_executions:
                if states.is_paused(task_ex.state):
                    return

            # Otherwise if no other subworkflow is paused,
            # then resume the parent workflow execution.
            wf_handler.resume_workflow(wf_ex)

    except exc.MistralException as e:
        wf_ex = task_ex.workflow_execution

        msg = ("Failed to handle action update [error=%s, wf=%s, task=%s,"
               " action=%s]:\n%s" %
               (e, wf_ex.name, task_ex.name, action_ex.name, tb.format_exc()))

        LOG.error(msg)

        task.set_state(states.ERROR, msg)

        wf_handler.force_fail_workflow(wf_ex, msg)

        return


def force_fail_task(task_ex, msg):
    """Forces the given task to fail.

    This method implements the 'forced' task fail without giving a chance
    to a workflow controller to handle the error. Its main purpose is to
    reflect errors caused by workflow structure (errors 'publish', 'on-xxx'
    clauses etc.) rather than failed actions. If such an error happens
    we should also force the entire workflow to fail. I.e., this kind of
    error must be propagated to a higher level, to the workflow.

    :param task_ex: Task execution.
    :param msg: Error message.
    """
    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )

    task = _build_task_from_execution(wf_spec, task_ex)

    task.set_state(states.ERROR, msg)

    wf_handler.force_fail_workflow(task_ex.workflow_execution, msg)


def continue_task(task_ex):
    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )

    task = _build_task_from_execution(wf_spec, task_ex)

    try:
        task.set_state(states.RUNNING, None)

        task.run()
    except exc.MistralException as e:
        wf_ex = task_ex.workflow_execution

        msg = (
            "Failed to run task [error=%s, wf=%s, task=%s]:\n%s" %
            (e, wf_ex.name, task_ex.name, tb.format_exc())
        )

        LOG.error(msg)

        task.set_state(states.ERROR, msg)

        wf_handler.force_fail_workflow(wf_ex, msg)

        return


def complete_task(task_ex, state, state_info):
    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )

    task = _build_task_from_execution(wf_spec, task_ex)

    try:
        task.complete(state, state_info)
    except exc.MistralException as e:
        wf_ex = task_ex.workflow_execution

        msg = (
            "Failed to complete task [error=%s, wf=%s, task=%s]:\n%s" %
            (e, wf_ex.name, task_ex.name, tb.format_exc())
        )

        LOG.error(msg)

        task.set_state(states.ERROR, msg)

        wf_handler.force_fail_workflow(wf_ex, msg)

        return


def _build_task_from_execution(wf_spec, task_ex):
    return _create_task(
        task_ex.workflow_execution,
        wf_spec,
        wf_spec.get_task(task_ex.name),
        task_ex.in_context,
        task_ex
    )


@profiler.trace('task-handler-build-task-from-command', hide_args=True)
def _build_task_from_command(cmd):
    if isinstance(cmd, wf_cmds.RunExistingTask):
        task = _create_task(
            cmd.wf_ex,
            cmd.wf_spec,
            spec_parser.get_task_spec(cmd.task_ex.spec),
            cmd.ctx,
            task_ex=cmd.task_ex,
            unique_key=cmd.task_ex.unique_key,
            waiting=cmd.task_ex.state == states.WAITING,
            triggered_by=cmd.triggered_by
        )

        if cmd.reset:
            task.reset()

        return task

    if isinstance(cmd, wf_cmds.RunTask):
        task = _create_task(
            cmd.wf_ex,
            cmd.wf_spec,
            cmd.task_spec,
            cmd.ctx,
            unique_key=cmd.unique_key,
            waiting=cmd.is_waiting(),
            triggered_by=cmd.triggered_by
        )

        return task

    raise exc.MistralError('Unsupported workflow command: %s' % cmd)


def _create_task(wf_ex, wf_spec, task_spec, ctx, task_ex=None,
                 unique_key=None, waiting=False, triggered_by=None):
    if task_spec.get_with_items():
        cls = tasks.WithItemsTask
    else:
        cls = tasks.RegularTask

    return cls(
        wf_ex,
        wf_spec,
        task_spec,
        ctx,
        task_ex=task_ex,
        unique_key=unique_key,
        waiting=waiting,
        triggered_by=triggered_by
    )


@db_utils.retry_on_db_error
@action_queue.process
@profiler.trace('task-handler-refresh-task-state', hide_args=True)
def _refresh_task_state(task_ex_id):
    with db_api.transaction():
        task_ex = db_api.load_task_execution(task_ex_id)

        if not task_ex:
            return

        wf_ex = task_ex.workflow_execution

        if states.is_completed(wf_ex.state):
            return

        wf_spec = spec_parser.get_workflow_spec_by_execution_id(
            task_ex.workflow_execution_id
        )

        wf_ctrl = wf_base.get_controller(wf_ex, wf_spec)

        log_state = wf_ctrl.get_logical_task_state(
            task_ex
        )

        state = log_state.state
        state_info = log_state.state_info

        # Update 'triggered_by' because it could have changed.
        task_ex.runtime_context['triggered_by'] = log_state.triggered_by

        if state == states.RUNNING:
            continue_task(task_ex)
        elif state == states.ERROR:
            complete_task(task_ex, state, state_info)
        elif state == states.WAITING:
            # Let's assume that a task takes 0.01 sec in average to complete
            # and based on this assumption calculate a time of the next check.
            # The estimation is very rough, of course, but this delay will be
            # decreasing as task preconditions will be completing which will
            # give a decent asymptotic approximation.
            # For example, if a 'join' task has 100 inbound incomplete tasks
            # then the next 'refresh_task_state' call will happen in 10
            # seconds. For 500 tasks it will be 50 seconds. The larger the
            # workflow is, the more beneficial this mechanism will be.
            delay = int(log_state.cardinality * 0.01)

            _schedule_refresh_task_state(task_ex, max(1, delay))
        else:
            # Must never get here.
            raise RuntimeError(
                'Unexpected logical task state [task_ex_id=%s, task_name=%s, '
                'state=%s]' % (task_ex_id, task_ex.name, state)
            )


def _schedule_refresh_task_state(task_ex, delay=0):
    """Schedules task preconditions check.

    This method provides transactional decoupling of task preconditions
    check from events that can potentially satisfy those preconditions.

    It's needed in non-locking model in order to avoid 'phantom read'
    phenomena when reading state of multiple tasks to see if a task that
    depends on them can start. Just starting a separate transaction
    without using scheduler is not safe due to concurrency window that
    we'll have in this case (time between transactions) whereas scheduler
    is a special component that is designed to be resistant to failures.

    :param task_ex: Task execution.
    :param delay: Delay.
    """
    key = 'th_c_t_s_a-%s' % task_ex.id

    scheduler.schedule_call(
        None,
        _REFRESH_TASK_STATE_PATH,
        delay,
        key=key,
        task_ex_id=task_ex.id
    )


@db_utils.retry_on_db_error
@action_queue.process
def _scheduled_on_action_complete(action_ex_id, wf_action):
    with db_api.transaction():
        if wf_action:
            action_ex = db_api.get_workflow_execution(action_ex_id)
        else:
            action_ex = db_api.get_action_execution(action_ex_id)

        _on_action_complete(action_ex)


def schedule_on_action_complete(action_ex, delay=0):
    """Schedules task completion check.

    This method provides transactional decoupling of action completion from
    task completion check. It's needed in non-locking model in order to
    avoid 'phantom read' phenomena when reading state of multiple actions
    to see if a task is completed. Just starting a separate transaction
    without using scheduler is not safe due to concurrency window that we'll
    have in this case (time between transactions) whereas scheduler is a
    special component that is designed to be resistant to failures.

    :param action_ex: Action execution.
    :param delay: Minimum amount of time before task completion check
        should be made.
    """

    # Optimization to avoid opening a new transaction if it's not needed.
    if not action_ex.task_execution.spec.get('with-items'):
        _on_action_complete(action_ex)

        return

    key = 'th_on_a_c-%s' % action_ex.task_execution_id

    scheduler.schedule_call(
        None,
        _SCHEDULED_ON_ACTION_COMPLETE_PATH,
        delay,
        key=key,
        action_ex_id=action_ex.id,
        wf_action=isinstance(action_ex, models.WorkflowExecution)
    )


@db_utils.retry_on_db_error
@action_queue.process
def _scheduled_on_action_update(action_ex_id, wf_action):
    with db_api.transaction():
        if wf_action:
            action_ex = db_api.get_workflow_execution(action_ex_id)
        else:
            action_ex = db_api.get_action_execution(action_ex_id)

        _on_action_update(action_ex)


def schedule_on_action_update(action_ex, delay=0):
    """Schedules task update check.

    This method provides transactional decoupling of action update from
    task update check. It's needed in non-locking model in order to
    avoid 'phantom read' phenomena when reading state of multiple actions
    to see if a task is updated. Just starting a separate transaction
    without using scheduler is not safe due to concurrency window that we'll
    have in this case (time between transactions) whereas scheduler is a
    special component that is designed to be resistant to failures.

    :param action_ex: Action execution.
    :param delay: Minimum amount of time before task update check
        should be made.
    """

    # Optimization to avoid opening a new transaction if it's not needed.
    if not action_ex.task_execution.spec.get('with-items'):
        _on_action_update(action_ex)

        return

    key = 'th_on_a_c-%s' % action_ex.task_execution_id

    scheduler.schedule_call(
        None,
        _SCHEDULED_ON_ACTION_UPDATE_PATH,
        delay,
        key=key,
        action_ex_id=action_ex.id,
        wf_action=isinstance(action_ex, models.WorkflowExecution)
    )



# Copyright 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Increase executions_v2 column size from JsonDictType to JsonLongDictType

Revision ID: 005
Revises: 004
Create Date: 2015-07-21 08:48:51.636094

"""

# revision identifiers, used by Alembic.
revision = '005'
down_revision = '004'

from alembic import op
from mistral.db.sqlalchemy import types as st


def upgrade():
    # Changing column types from JsonDictType to JsonLongDictType
    op.alter_column('executions_v2', 'runtime_context',
                    type_=st.JsonLongDictType())
    op.alter_column('executions_v2', 'input',
                    type_=st.JsonLongDictType())
    op.alter_column('executions_v2', 'params',
                    type_=st.JsonLongDictType())
    op.alter_column('executions_v2', 'context',
                    type_=st.JsonLongDictType())
    op.alter_column('executions_v2', 'action_spec',
                    type_=st.JsonLongDictType())
    op.alter_column('executions_v2', 'published',
                    type_=st.JsonLongDictType())



# Copyright 2015 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from oslo_config import cfg

from mistral.actions import std_actions
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral.services import actions
from mistral.tests.unit.engine import base
from mistral.workflow import states
from mistral_lib import actions as ml_actions

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class RunActionEngineTest(base.EngineTestCase):
    @classmethod
    def heavy_init(cls):
        super(RunActionEngineTest, cls).heavy_init()

        action = """---
        version: '2.0'

        concat:
          base: std.echo
          base-input:
            output: <% $.left %><% $.right %>
          input:
            - left
            - right

        concat3:
          base: concat
          base-input:
            left: <% $.left %><% $.center %>
            right: <% $.right %>
          input:
            - left
            - center
            - right

        concat4:
          base: concat3
          base-input:
            left: <% $.left %>
            center: <% $.center_left %><% $.center_right %>
            right: <% $.right %>
          input:
            - left
            - center_left
            - center_right
            - right

        missing_base:
          base: wrong
          input:
            - some_input

        nested_missing_base:
          base: missing_base
          input:
            - some_input

        loop_action:
          base: loop_action
          base-input:
            output: <% $.output %>
          input:
            - output

        level2_loop_action:
          base: loop_action
          base-input:
            output: <% $.output %>
          input:
            - output
        """
        actions.create_actions(action)

    def test_run_action_sync(self):
        # Start action and see the result.
        action_ex = self.engine.start_action('std.echo', {'output': 'Hello!'})

        self.assertEqual('Hello!', action_ex.output['result'])
        self.assertEqual(states.SUCCESS, action_ex.state)

    @mock.patch.object(
        std_actions.EchoAction,
        'run',
        mock.Mock(side_effect=exc.ActionException("some error"))
    )
    def test_run_action_error(self):
        # Start action and see the result.
        action_ex = self.engine.start_action('std.echo', {'output': 'Hello!'})

        self.assertIsNotNone(action_ex.output)
        self.assertIn('some error', action_ex.output['result'])
        self.assertEqual(states.ERROR, action_ex.state)

    def test_run_action_save_result(self):
        # Start action.
        action_ex = self.engine.start_action(
            'std.echo',
            {'output': 'Hello!'},
            save_result=True
        )

        self.await_action_success(action_ex.id)

        with db_api.transaction():
            action_ex = db_api.get_action_execution(action_ex.id)

            self.assertEqual(states.SUCCESS, action_ex.state)
            self.assertEqual({'result': 'Hello!'}, action_ex.output)

    def test_run_action_run_sync(self):
        # Start action.
        action_ex = self.engine.start_action(
            'std.echo',
            {'output': 'Hello!'},
            run_sync=True
        )

        self.assertEqual('Hello!', action_ex.output['result'])
        self.assertEqual(states.SUCCESS, action_ex.state)

    def test_run_action_save_result_and_run_sync(self):
        # Start action.
        action_ex = self.engine.start_action(
            'std.echo',
            {'output': 'Hello!'},
            save_result=True,
            run_sync=True
        )

        self.assertEqual('Hello!', action_ex.output['result'])
        self.assertEqual(states.SUCCESS, action_ex.state)

        with db_api.transaction():
            action_ex = db_api.get_action_execution(action_ex.id)

            self.assertEqual(states.SUCCESS, action_ex.state)
            self.assertEqual({'result': 'Hello!'}, action_ex.output)

    def test_run_action_run_sync_error(self):
        # Start action.
        self.assertRaises(
            exc.InputException,
            self.engine.start_action,
            'std.async_noop',
            {},
            run_sync=True
        )

    def test_run_action_async(self):
        action_ex = self.engine.start_action('std.async_noop', {})

        self.await_action_state(action_ex.id, states.RUNNING)

        action_ex = db_api.get_action_execution(action_ex.id)

        self.assertEqual(states.RUNNING, action_ex.state)

    @mock.patch.object(
        std_actions.AsyncNoOpAction, 'run',
        mock.MagicMock(side_effect=exc.ActionException('Invoke failed.')))
    def test_run_action_async_invoke_failure(self):
        action_ex = self.engine.start_action('std.async_noop', {})

        self.await_action_error(action_ex.id)

        with db_api.transaction():
            action_ex = db_api.get_action_execution(action_ex.id)

            self.assertEqual(states.ERROR, action_ex.state)
            self.assertIn('Invoke failed.', action_ex.output.get('result', ''))

    @mock.patch.object(
        std_actions.AsyncNoOpAction, 'run',
        mock.MagicMock(return_value=ml_actions.Result(error='Invoke erred.')))
    def test_run_action_async_invoke_with_error(self):
        action_ex = self.engine.start_action('std.async_noop', {})

        self.await_action_error(action_ex.id)

        with db_api.transaction():
            action_ex = db_api.get_action_execution(action_ex.id)

            self.assertEqual(states.ERROR, action_ex.state)
            self.assertIn('Invoke erred.', action_ex.output.get('result', ''))

    def test_run_action_adhoc(self):
        # Start action and see the result.
        action_ex = self.engine.start_action(
            'concat',
            {'left': 'Hello, ', 'right': 'John Doe!'}
        )

        self.assertEqual('Hello, John Doe!', action_ex.output['result'])

    def test_run_level_two_action_adhoc(self):
        # Start action and see the result.
        action_ex = self.engine.start_action(
            'concat3',
            {'left': 'Hello, ', 'center': 'John', 'right': ' Doe!'}
        )

        self.assertEqual('Hello, John Doe!', action_ex.output['result'])

    def test_run_level_three_action_adhoc(self):
        # Start action and see the result.
        action_ex = self.engine.start_action(
            'concat4',
            {
                'left': 'Hello, ',
                'center_left': 'John',
                'center_right': ' Doe',
                'right': '!'
            }
        )

        self.assertEqual('Hello, John Doe!', action_ex.output['result'])

    def test_run_action_with_missing_base(self):
        # Start action and see the result.
        self.assertRaises(
            exc.InvalidActionException,
            self.engine.start_action,
            'missing_base',
            {'some_input': 'Hi'}
        )

    def test_run_action_with_missing_nested_base(self):
        # Start action and see the result.
        self.assertRaises(
            exc.InvalidActionException,
            self.engine.start_action,
            'nested_missing_base',
            {'some_input': 'Hi'}
        )

    def test_run_loop_action(self):
        # Start action and see the result.
        self.assertRaises(
            ValueError,
            self.engine.start_action,
            'loop_action',
            {'output': 'Hello'}
        )

    def test_run_level_two_loop_action(self):
        # Start action and see the result.
        self.assertRaises(
            ValueError,
            self.engine.start_action,
            'level2_loop_action',
            {'output': 'Hello'}
        )

    def test_run_action_wrong_input(self):
        # Start action and see the result.
        exception = self.assertRaises(
            exc.InputException,
            self.engine.start_action,
            'std.http',
            {'url': 'Hello, ', 'metod': 'John Doe!'}
        )

        self.assertIn('std.http', str(exception))

    def test_adhoc_action_wrong_input(self):
        # Start action and see the result.
        exception = self.assertRaises(
            exc.InputException,
            self.engine.start_action,
            'concat',
            {'left': 'Hello, ', 'ri': 'John Doe!'}
        )

        self.assertIn('concat', str(exception))

    # TODO(rakhmerov): This is an example of a bad test. It pins to
    # implementation details too much and prevents from making refactoring
    # easily. When writing tests we should make assertions about
    # consequences, not about how internal machinery works, i.e. we need to
    # follow "black box" testing paradigm.
    @mock.patch('mistral.engine.actions.resolve_action_definition')
    @mock.patch('mistral.engine.utils.validate_input')
    @mock.patch('mistral.services.action_manager.get_action_class')
    @mock.patch('mistral.engine.actions.PythonAction.run')
    def test_run_action_with_kwargs_input(self, run_mock, class_mock,
                                          validate_mock, def_mock):
        action_def = models.ActionDefinition()
        action_def.update({
            'name': 'fake_action',
            'action_class': '',
            'attributes': {},
            'description': '',
            'input': '**kwargs',
            'is_system': True,
            'scope': 'public'
        })
        def_mock.return_value = action_def
        run_mock.return_value = ml_actions.Result(data='Hello')

        class_ret = mock.MagicMock()
        class_mock.return_value = class_ret

        self.engine.start_action('fake_action', {'input': 'Hello'})

        self.assertEqual(1, def_mock.call_count)
        def_mock.assert_called_with('fake_action')

        self.assertEqual(0, validate_mock.call_count)

        class_ret.assert_called_once_with(input='Hello')

        run_mock.assert_called_once_with(
            {'input': 'Hello'},
            None,
            save=False,
            timeout=None
        )



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import mock
import pecan
import pecan.testing
import requests
import requests_mock
import webob

from mistral.api import app as pecan_app
from mistral.auth import keycloak
from mistral import context
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral.services import periodic
from mistral.tests.unit import base
from mistral.tests.unit.mstrlfixtures import policy_fixtures


WF_DEFINITION = """
---
version: '2.0'
flow:
  type: direct
  input:
    - param1
  tasks:
    task1:
      action: std.echo output="Hi"
"""

WF_DB = models.WorkflowDefinition(
    id='123e4567-e89b-12d3-a456-426655440000',
    name='flow',
    definition=WF_DEFINITION,
    created_at=datetime.datetime(1970, 1, 1),
    updated_at=datetime.datetime(1970, 1, 1),
    spec={'input': ['param1']}
)

WF = {
    'id': '123e4567-e89b-12d3-a456-426655440000',
    'name': 'flow',
    'definition': WF_DEFINITION,
    'created_at': '1970-01-01 00:00:00',
    'updated_at': '1970-01-01 00:00:00',
    'input': 'param1'
}


MOCK_WF = mock.MagicMock(return_value=WF_DB)

# Set up config options.
AUTH_URL = 'https://my.keycloak.com:8443/auth'
REALM_NAME = 'my_realm'

USER_INFO_ENDPOINT = (
    "%s/realms/%s/protocol/openid-connect/userinfo" % (AUTH_URL, REALM_NAME)
)

USER_CLAIMS = {
    "sub": "248289761001",
    "name": "Jane Doe",
    "given_name": "Jane",
    "family_name": "Doe",
    "preferred_username": "j.doe",
    "email": "janedoe@example.com",
    "picture": "http://example.com/janedoe/me.jpg"
}

WWW_AUTHENTICATE_HEADER = {'WWW-Authenticate': 'unauthorized reason is ...'}


class TestKeyCloakOIDCAuth(base.BaseTest):

    def setUp(self):
        super(TestKeyCloakOIDCAuth, self).setUp()

        self.override_config('auth_url', AUTH_URL, group='keycloak_oidc')

        self.auth_handler = keycloak.KeycloakAuthHandler()

    def _build_request(self, token):
        req = webob.Request.blank("/")

        req.headers["x-auth-token"] = token
        req.get_response = lambda app: None

        return req

    @requests_mock.Mocker()
    def test_header_parsing(self, req_mock):
        token = {
            "iss": "http://localhost:8080/auth/realms/my_realm",
            "realm_access": {
                "roles": ["role1", "role2"]
            }
        }

        # Imitate successful response from KeyCloak with user claims.
        req_mock.get(USER_INFO_ENDPOINT, json=USER_CLAIMS)

        req = self._build_request(token)

        with mock.patch("jwt.decode", return_value=token):
            self.auth_handler.authenticate(req)

        self.assertEqual("Confirmed", req.headers["X-Identity-Status"])
        self.assertEqual("my_realm", req.headers["X-Project-Id"])
        self.assertEqual("role1,role2", req.headers["X-Roles"])
        self.assertEqual(1, req_mock.call_count)

    def test_no_auth_token(self):
        req = webob.Request.blank("/")

        self.assertRaises(
            exc.UnauthorizedException,
            self.auth_handler.authenticate,
            req
        )

    @requests_mock.Mocker()
    def test_no_realm_roles(self, req_mock):
        token = {"iss": "http://localhost:8080/auth/realms/my_realm"}

        # Imitate successful response from KeyCloak with user claims.
        req_mock.get(USER_INFO_ENDPOINT, json=USER_CLAIMS)

        req = self._build_request(token)

        with mock.patch("jwt.decode", return_value=token):
            self.auth_handler.authenticate(req)

        self.assertEqual("Confirmed", req.headers["X-Identity-Status"])
        self.assertEqual("my_realm", req.headers["X-Project-Id"])
        self.assertEqual("", req.headers["X-Roles"])

    def test_wrong_token_format(self):
        req = self._build_request(token="WRONG_FORMAT_TOKEN")

        self.assertRaises(
            exc.UnauthorizedException,
            self.auth_handler.authenticate,
            req
        )

    @requests_mock.Mocker()
    def test_server_unauthorized(self, req_mock):
        token = {
            "iss": "http://localhost:8080/auth/realms/my_realm",
        }

        # Imitate failure response from KeyCloak.
        req_mock.get(
            USER_INFO_ENDPOINT,
            status_code=401,
            reason='Access token is invalid',
            headers=WWW_AUTHENTICATE_HEADER
        )

        req = self._build_request(token)

        with mock.patch("jwt.decode", return_value=token):
            try:
                self.auth_handler.authenticate(req)
            except requests.exceptions.HTTPError as e:
                self.assertIn(
                    "401 Client Error: Access token is invalid for url",
                    str(e)
                )
                self.assertEqual(
                    'unauthorized reason is ...',
                    e.response.headers.get('WWW-Authenticate')
                )

            else:
                raise Exception("Test is broken")

    @requests_mock.Mocker()
    def test_connection_error(self, req_mock):
        token = {"iss": "http://localhost:8080/auth/realms/my_realm"}

        req_mock.get(USER_INFO_ENDPOINT, exc=requests.ConnectionError)

        req = self._build_request(token)

        with mock.patch("jwt.decode", return_value=token):
            self.assertRaises(
                exc.MistralException,
                self.auth_handler.authenticate,
                req
            )


class TestKeyCloakOIDCAuthScenarios(base.DbTestCase):
    def setUp(self):
        super(TestKeyCloakOIDCAuthScenarios, self).setUp()

        self.override_config('enabled', False, group='cron_trigger')
        self.override_config('auth_enable', True, group='pecan')
        self.override_config('auth_type', 'keycloak-oidc')
        self.override_config('auth_url', AUTH_URL, group='keycloak_oidc')

        self.app = pecan.testing.load_test_app(
            dict(pecan_app.get_pecan_config())
        )

        # Adding cron trigger thread clean up explicitly in case if
        # new tests will provide an alternative configuration for pecan
        # application.
        self.addCleanup(periodic.stop_all_periodic_tasks)

        # Make sure the api get the correct context.
        self.patch_ctx = mock.patch(
            'mistral.context.MistralContext.from_environ'
        )
        self.mock_ctx = self.patch_ctx.start()
        self.mock_ctx.return_value = self.ctx

        self.addCleanup(self.patch_ctx.stop)

        self.policy = self.useFixture(policy_fixtures.PolicyFixture())

    @requests_mock.Mocker()
    @mock.patch.object(db_api, 'get_workflow_definition', MOCK_WF)
    def test_get_workflow_success_auth(self, req_mock):
        # Imitate successful response from KeyCloak with user claims.
        req_mock.get(USER_INFO_ENDPOINT, json=USER_CLAIMS)

        token = {
            "iss": "http://localhost:8080/auth/realms/%s" % REALM_NAME,
            "realm_access": {
                "roles": ["role1", "role2"]
            }
        }

        headers = {'X-Auth-Token': str(token)}

        with mock.patch("jwt.decode", return_value=token):
            resp = self.app.get('/v2/workflows/123', headers=headers)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(WF, resp.json)

    @mock.patch("requests.get")
    @mock.patch.object(db_api, 'get_workflow_definition', MOCK_WF)
    def test_get_workflow_invalid_token_format(self, req_mock):
        # Imitate successful response from KeyCloak with user claims.
        req_mock.get(USER_INFO_ENDPOINT, json=USER_CLAIMS)

        token = {
            "iss": "http://localhost:8080/auth/realms/%s" % REALM_NAME,
            "realm_access": {
                "roles": ["role1", "role2"]
            }
        }

        headers = {'X-Auth-Token': str(token)}

        resp = self.app.get(
            '/v2/workflows/123',
            headers=headers,
            expect_errors=True
        )

        self.assertEqual(401, resp.status_code)
        self.assertEqual('401 Unauthorized', resp.status)
        self.assertIn('Failed to validate access token', resp.text)
        self.assertIn(
            "Token can't be decoded because of wrong format",
            resp.text
        )

    @requests_mock.Mocker()
    @mock.patch.object(db_api, 'get_workflow_definition', MOCK_WF)
    def test_get_workflow_failed_auth(self, req_mock):
        # Imitate failure response from KeyCloak.
        req_mock.get(
            USER_INFO_ENDPOINT,
            status_code=401,
            reason='Access token is invalid'
        )

        token = {
            "iss": "http://localhost:8080/auth/realms/%s" % REALM_NAME,
            "realm_access": {
                "roles": ["role1", "role2"]
            }
        }

        headers = {'X-Auth-Token': str(token)}

        with mock.patch("jwt.decode", return_value=token):
            resp = self.app.get(
                '/v2/workflows/123',
                headers=headers,
                expect_errors=True
            )

        self.assertEqual(401, resp.status_code)
        self.assertEqual('401 Unauthorized', resp.status)
        self.assertIn('Failed to validate access token', resp.text)
        self.assertIn('Access token is invalid', resp.text)


class TestKeyCloakOIDCAuthApp(base.DbTestCase):
    """Test that Keycloak auth params were successfully passed to Context"""

    def setUp(self):
        super(TestKeyCloakOIDCAuthApp, self).setUp()

        self.override_config('enabled', False, group='cron_trigger')
        self.override_config('auth_enable', True, group='pecan')
        self.override_config('auth_type', 'keycloak-oidc')
        self.override_config('auth_url', AUTH_URL, group='keycloak_oidc')

        self.app = pecan.testing.load_test_app(
            dict(pecan_app.get_pecan_config())
        )

        # Adding cron trigger thread clean up explicitly in case if
        # new tests will provide an alternative configuration for pecan
        # application.
        self.addCleanup(periodic.stop_all_periodic_tasks)

        self.policy = self.useFixture(policy_fixtures.PolicyFixture())

    @requests_mock.Mocker()
    @mock.patch.object(db_api, 'get_workflow_definition', MOCK_WF)
    def test_params_transition(self, req_mock):
        req_mock.get(USER_INFO_ENDPOINT, json=USER_CLAIMS)

        token = {
            "iss": "http://localhost:8080/auth/realms/%s" % REALM_NAME,
            "realm_access": {
                "roles": ["role1", "role2"]
            }
        }

        headers = {
            'X-Auth-Token': str(token)
        }

        with mock.patch("jwt.decode", return_value=token):
            with mock.patch("mistral.context.set_ctx") as mocked_set_cxt:
                self.app.get('/v2/workflows/123', headers=headers)
                calls = mocked_set_cxt.call_args_list
                self.assertEqual(2, len(calls))

                # First positional argument of the first call ('before')
                ctx = calls[0][0][0]

                self.assertIsInstance(ctx, context.MistralContext)
                self.assertEqual('my_realm', ctx.project_id)
                self.assertEqual(["role1", "role2"], ctx.roles)

                # Second call of set_ctx ('after'), where we reset the context
                self.assertIsNone(calls[1][0][0])



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - Huawei Technologies Co. Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy

from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral import utils


def _compare_parameters(expected_input, actual_input):
    """Compares the expected parameters with the actual parameters.

    :param expected_input: Expected dict of parameters.
    :param actual_input: Actual dict of parameters.
    :return: Tuple {missing parameter names, unexpected parameter names}
    """

    missing_params = []
    unexpected_params = copy.deepcopy(list((actual_input or {}).keys()))

    for p_name, p_value in expected_input.items():
        if p_value is utils.NotDefined and p_name not in unexpected_params:
            missing_params.append(str(p_name))

        if p_name in unexpected_params:
            unexpected_params.remove(p_name)

    return missing_params, unexpected_params


def validate_input(expected_input, actual_input, obj_name, obj_class):
    actual_input = actual_input or {}

    missing, unexpected = _compare_parameters(
        expected_input,
        actual_input
    )

    if missing or unexpected:
        msg = 'Invalid input [name=%s, class=%s'
        msg_props = [obj_name, obj_class]

        if missing:
            msg += ', missing=%s'
            msg_props.append(missing)

        if unexpected:
            msg += ', unexpected=%s'
            msg_props.append(unexpected)

        msg += ']'

        raise exc.InputException(msg % tuple(msg_props))


def resolve_workflow_definition(parent_wf_name, parent_wf_spec_name,
                                namespace, wf_spec_name):
    wf_def = None

    if parent_wf_name != parent_wf_spec_name:
        # If parent workflow belongs to a workbook then
        # check child workflow within the same workbook
        # (to be able to use short names within workbooks).
        # If it doesn't exist then use a name from spec
        # to find a workflow in DB.
        wb_name = parent_wf_name.rstrip(parent_wf_spec_name)[:-1]

        wf_full_name = "%s.%s" % (wb_name, wf_spec_name)

        wf_def = db_api.load_workflow_definition(wf_full_name, namespace)

    if not wf_def:
        wf_def = db_api.load_workflow_definition(wf_spec_name, namespace)

    if not wf_def:
        raise exc.WorkflowException(
            "Failed to find workflow [name=%s] [namespace=%s]" %
            (wf_spec_name, namespace)
        )

    return wf_def



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.services import workbooks as wb_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WORKBOOK1 = """
---
version: '2.0'

name: my_wb

workflows:
  wf:
    type: direct
    input:
      - my_var

    tasks:
      task1:
        action: std.echo output='1'
        on-complete:
          - fail: <% $.my_var = 1 %>
          - succeed: <% $.my_var = 2 %>
          - pause: <% $.my_var = 3 %>
          - task2

      task2:
        action: std.echo output='2'
"""


class SimpleEngineCommandsTest(base.EngineTestCase):
    def setUp(self):
        super(SimpleEngineCommandsTest, self).setUp()

        wb_service.create_workbook_v2(WORKBOOK1)

    def test_fail(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

    def test_succeed(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

    def test_pause(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})

        self.await_workflow_paused(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

        # Let's resume the workflow and wait till it succeeds.
        self.engine.resume_workflow(wf_ex.id)

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self._assert_single_item(
            task_execs,
            name='task2',
            state=states.SUCCESS
        )


WORKBOOK2 = """
---
version: '2.0'

name: my_wb

workflows:
  wf:
    type: direct
    input:
      - my_var

    task-defaults:
      on-complete:
        - fail: <% $.my_var = 1 %>
        - succeed: <% $.my_var = 2 %>
        - pause: <% $.my_var = 3 %>
        - task2: <% $.my_var = 4 %> # (Never happens in this test)

    tasks:
      task1:
        action: std.echo output='1'

      task2:
        action: std.echo output='2'
"""


class SimpleEngineWorkflowLevelCommandsTest(base.EngineTestCase):
    def setUp(self):
        super(SimpleEngineWorkflowLevelCommandsTest, self).setUp()

        wb_service.create_workbook_v2(WORKBOOK2)

    def test_fail(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

    def test_succeed(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

    def test_pause(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})

        self.await_workflow_paused(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )


WORKBOOK3 = """
---
version: '2.0'

name: my_wb

workflows:
  fail_first_wf:
    type: direct

    tasks:
      task1:
        action: std.echo output='1'
        on-complete:
          - fail
          - task2

      task2:
        action: std.echo output='2'

  fail_second_wf:
    type: direct

    tasks:
      task1:
        action: std.echo output='1'
        on-complete:
          - task2
          - fail

      task2:
        action: std.echo output='2'

  succeed_first_wf:
    type: direct

    tasks:
      task1:
        action: std.echo output='1'
        on-complete:
          - succeed
          - task2

      task2:
        action: std.echo output='2'

  succeed_second_wf:
    type: direct

    tasks:
      task1:
        action: std.echo output='1'
        on-complete:
          - task2
          - succeed

      task2:
        action: std.http url='some.not.existing.url'
"""


class OrderEngineCommandsTest(base.EngineTestCase):
    def setUp(self):
        super(OrderEngineCommandsTest, self).setUp()

        wb_service.create_workbook_v2(WORKBOOK3)

    def test_fail_first(self):
        wf_ex = self.engine.start_workflow('my_wb.fail_first_wf')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

    def test_fail_second(self):
        wf_ex = self.engine.start_workflow('my_wb.fail_second_wf')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        task2_db = self._assert_single_item(task_execs, name='task2')

        self.await_task_success(task2_db.id)
        self.await_workflow_error(wf_ex.id)

    def test_succeed_first(self):
        wf_ex = self.engine.start_workflow('my_wb.succeed_first_wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

    def test_succeed_second(self):
        wf_ex = self.engine.start_workflow('my_wb.succeed_second_wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        task2_db = self._assert_single_item(task_execs, name='task2')

        self.await_task_error(task2_db.id)
        self.await_workflow_success(wf_ex.id)

WORKBOOK4 = """
---
version: '2.0'

name: my_wb

workflows:
  wf:
    type: direct
    input:
      - my_var

    tasks:
      task1:
        action: std.echo output='1'
        on-complete:
        - fail(msg='my_var value is 1'): <% $.my_var = 1 %>
        - succeed(msg='my_var value is 2'): <% $.my_var = 2 %>
        - pause(msg='my_var value is 3'): <% $.my_var = 3 %>
        - task2

      task2:
        action: std.echo output='2'
"""


class SimpleEngineCmdsWithMsgTest(base.EngineTestCase):
    def setUp(self):
        super(SimpleEngineCmdsWithMsgTest, self).setUp()

        wb_service.create_workbook_v2(WORKBOOK4)

    def test_fail(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self.assertEqual(states.ERROR, wf_ex.state)
        self.assertEqual('my_var value is 1', wf_ex.state_info)

    def test_succeed(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self.assertEqual(states.SUCCESS, wf_ex.state)
        self.assertEqual("my_var value is 2", wf_ex.state_info)

    def test_pause(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})

        self.await_workflow_paused(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self.assertEqual(states.PAUSED, wf_ex.state)
        self.assertEqual("my_var value is 3", wf_ex.state_info)

WORKBOOK5 = """
---
version: '2.0'

name: my_wb

workflows:
  wf:
    type: direct
    input:
      - my_var

    task-defaults:
      on-complete:
        - fail(msg='my_var value is 1'): <% $.my_var = 1 %>
        - succeed(msg='my_var value is <% $.my_var %>'): <% $.my_var = 2 %>
        - pause(msg='my_var value is 3'): <% $.my_var = 3 %>
        - task2: <% $.my_var = 4 %> # (Never happens in this test)

    tasks:
      task1:
        action: std.echo output='1'

      task2:
        action: std.echo output='2'
"""


class SimpleEngineWorkflowLevelCmdsWithMsgTest(base.EngineTestCase):
    def setUp(self):
        super(SimpleEngineWorkflowLevelCmdsWithMsgTest, self).setUp()

        wb_service.create_workbook_v2(WORKBOOK5)

    def test_fail(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self.assertEqual(states.ERROR, wf_ex.state)
        self.assertEqual("my_var value is 1", wf_ex.state_info)

    def test_succeed(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self.assertEqual(states.SUCCESS, wf_ex.state)
        self.assertEqual("my_var value is 2", wf_ex.state_info)

    def test_pause(self):
        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})

        self.await_workflow_paused(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )
        self.assertEqual(states.PAUSED, wf_ex.state)
        self.assertEqual("my_var value is 3", wf_ex.state_info)



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_db.sqlalchemy import models as oslo_models
import sqlalchemy as sa
from sqlalchemy import event
from sqlalchemy.ext import declarative
from sqlalchemy.orm import attributes

from mistral.services import security
from mistral import utils


def id_column():
    return sa.Column(
        sa.String(36),
        primary_key=True,
        default=utils.generate_unicode_uuid
    )


class _MistralModelBase(oslo_models.ModelBase, oslo_models.TimestampMixin):
    """Base class for all Mistral SQLAlchemy DB Models."""

    created_at = sa.Column(sa.DateTime, default=lambda: utils.utc_now_sec())
    updated_at = sa.Column(sa.DateTime, onupdate=lambda: utils.utc_now_sec())

    __table__ = None

    __hash__ = object.__hash__

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __eq__(self, other):
        if type(self) is not type(other):
            return False

        for col in self.__table__.columns:
            # In case of single table inheritance a class attribute
            # corresponding to a table column may not exist so we need
            # to skip these attributes.
            if (hasattr(self, col.name)
                and hasattr(other, col.name)
                    and getattr(self, col.name) != getattr(other, col.name)):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self):
        """sqlalchemy based automatic to_dict method."""

        d = {col_name: col_val for col_name, col_val in self.iter_columns()}

        utils.datetime_to_str_in_dict(d, 'created_at')
        utils.datetime_to_str_in_dict(d, 'updated_at')

        return d

    def iter_column_names(self):
        """Returns an iterator for loaded column names.

        :return: A generator function for column names.
        """

        # If a column is unloaded at this point, it is
        # probably deferred. We do not want to access it
        # here and thereby cause it to load.
        unloaded = attributes.instance_state(self).unloaded

        for col in self.__table__.columns:
            if col.name not in unloaded and hasattr(self, col.name):
                yield col.name

    def iter_columns(self):
        """Returns an iterator for loaded columns.

        :return: A generator function that generates
            tuples (column name, column value).
        """

        for col_name in self.iter_column_names():
            yield col_name, getattr(self, col_name)

    def get_clone(self):
        """Clones current object, loads all fields and returns the result."""
        m = self.__class__()

        for col in self.__table__.columns:
            if hasattr(self, col.name):
                setattr(m, col.name, getattr(self, col.name))

        setattr(
            m,
            'created_at',
            utils.datetime_to_str(getattr(self, 'created_at'))
        )

        updated_at = getattr(self, 'updated_at')

        # NOTE(nmakhotkin): 'updated_at' field is empty for just created
        # object since it has not updated yet.
        if updated_at:
            setattr(m, 'updated_at', utils.datetime_to_str(updated_at))

        return m

    def __repr__(self):
        return '%s %s' % (type(self).__name__, self.to_dict().__repr__())


MistralModelBase = declarative.declarative_base(cls=_MistralModelBase)


# Secure model related stuff.


class MistralSecureModelBase(MistralModelBase):
    """Base class for all secure models."""

    __abstract__ = True

    scope = sa.Column(sa.String(80), default='private')
    project_id = sa.Column(sa.String(80), default=security.get_project_id)


def _set_project_id(target, value, oldvalue, initiator):
    return security.get_project_id()


def register_secure_model_hooks():
    # Make sure 'project_id' is always properly set.
    for sec_model_class in utils.iter_subclasses(MistralSecureModelBase):
        if '__abstract__' not in sec_model_class.__dict__:
            event.listen(
                sec_model_class.project_id,
                'set',
                _set_project_id,
                retval=True
            )



# Copyright 2016 - Nokia Networks.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import eventlet
import functools

from oslo_config import cfg

from mistral import context
from mistral.executors import base as exe
from mistral.rpc import clients as rpc
from mistral import utils


_THREAD_LOCAL_NAME = "__action_queue_thread_local"

# Action queue operations.
_RUN_ACTION = "run_action"
_ON_ACTION_COMPLETE = "on_action_complete"


def _prepare():
    utils.set_thread_local(_THREAD_LOCAL_NAME, list())


def _clear():
    utils.set_thread_local(_THREAD_LOCAL_NAME, None)


def _get_queue():
    queue = utils.get_thread_local(_THREAD_LOCAL_NAME)

    if queue is None:
        raise RuntimeError(
            'Action queue is not initialized for the current thread.'
            ' Most likely some transactional method is not decorated'
            ' with action_queue.process()'
        )

    return queue


def _process_queue(queue):
    executor = exe.get_executor(cfg.CONF.executor.type)

    for operation, args in queue:
        if operation == _RUN_ACTION:
            action_ex, action_def, target, execution_context, timeout = args

            executor.run_action(
                action_ex.id,
                action_def.action_class,
                action_def.attributes or {},
                action_ex.input,
                action_ex.runtime_context.get('safe_rerun', False),
                execution_context,
                target=target,
                timeout=timeout
            )
        elif operation == _ON_ACTION_COMPLETE:
            action_ex_id, result, wf_action = args

            rpc.get_engine_client().on_action_complete(
                action_ex_id,
                result,
                wf_action
            )


def process(func):
    """Decorator that processes (runs) all actions in the action queue.

    Various engine methods may cause new actions to be scheduled. All
    such methods must be decorated with this decorator. It makes sure
    to run all the actions in the queue and clean up the queue.
    """
    @functools.wraps(func)
    def decorate(*args, **kw):
        _prepare()

        try:
            res = func(*args, **kw)

            queue = _get_queue()
            auth_ctx = context.ctx() if context.has_ctx() else None

            # NOTE(rakhmerov): Since we make RPC calls to the engine itself
            # we need to process the action queue asynchronously in a new
            # thread. Otherwise, if we have one engine process the engine
            # will may send a request to itself while already processing
            # another one. In conjunction with blocking RPC it will lead
            # to a deadlock (and RPC timeout).
            def _within_new_thread():
                old_auth_ctx = context.ctx() if context.has_ctx() else None

                context.set_ctx(auth_ctx)

                try:
                    _process_queue(queue)
                finally:
                    context.set_ctx(old_auth_ctx)

            eventlet.spawn(_within_new_thread)
        finally:
            _clear()

        return res

    return decorate


def schedule_run_action(action_ex, action_def, target, execution_context,
                        timeout):
    args = (action_ex, action_def, target, execution_context, timeout)
    _get_queue().append((_RUN_ACTION, args))


def schedule_on_action_complete(action_ex_id, result, wf_action=False):
    _get_queue().append(
        (_ON_ACTION_COMPLETE, (action_ex_id, result, wf_action))
    )



# Copyright 2014 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import eventlet
import mock

from oslo_config import cfg

from mistral import exceptions as exc
from mistral.rpc import clients as rpc
from mistral.services import periodic
from mistral.services import security
from mistral.services import triggers as t_s
from mistral.services import workflows
from mistral.tests.unit import base
from mistral import utils

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WORKFLOW_LIST = """
---
version: '2.0'

my_wf:
  type: direct

  tasks:
    task1:
      action: std.echo output='Hi!'
"""

advance_cron_trigger_orig = periodic.advance_cron_trigger


def new_advance_cron_trigger(ct):
    """Wrap the original advance_cron_trigger method.

    This method makes sure that the other coroutines will also run
    while this thread is executing. Without explicitly passing control to
    another coroutine the process_cron_triggers_v2 will finish looping
    over all the cron triggers in one coroutine without any sharing at all.
    """
    eventlet.sleep()
    modified = advance_cron_trigger_orig(ct)
    eventlet.sleep()

    return modified


class TriggerServiceV2Test(base.DbTestCase):
    def setUp(self):
        super(TriggerServiceV2Test, self).setUp()

        self.wf = workflows.create_workflows(WORKFLOW_LIST)[0]

    def test_trigger_create(self):
        trigger = t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '*/5 * * * *',
            None,
            None,
            datetime.datetime(2010, 8, 25)
        )

        self.assertEqual(
            datetime.datetime(2010, 8, 25, 0, 5),
            trigger.next_execution_time
        )

        next_time = t_s.get_next_execution_time(
            trigger['pattern'],
            trigger.next_execution_time
        )

        self.assertEqual(datetime.datetime(2010, 8, 25, 0, 10), next_time)

    def test_trigger_create_with_wf_id(self):
        trigger = t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            None,
            {},
            {},
            '*/5 * * * *',
            None,
            None,
            datetime.datetime(2010, 8, 25),
            workflow_id=self.wf.id
        )

        self.assertEqual(self.wf.name, trigger.workflow_name)

    def test_trigger_create_the_same_first_time_or_count(self):
        t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '*/5 * * * *',
            "4242-12-25 13:37",
            2,
            datetime.datetime(2010, 8, 25)
        )

        t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '*/5 * * * *',
            "4242-12-25 13:37",
            4,
            datetime.datetime(2010, 8, 25)
        )

        t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '*/5 * * * *',
            "5353-12-25 13:37",
            2,
            datetime.datetime(2010, 8, 25)
        )

        # Creations above should be ok.

        # But creation with the same count and first time
        # simultaneously leads to error.
        self.assertRaises(
            exc.DBDuplicateEntryError,
            t_s.create_cron_trigger,
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '*/5 * * * *',
            "4242-12-25 13:37",
            2,
            None
        )

    def test_trigger_create_wrong_workflow_input(self):
        wf_with_input = """---
        version: '2.0'

        some_wf:
          input:
            - some_var
          tasks:
            some_task:
              action: std.echo output=<% $.some_var %>
        """
        workflows.create_workflows(wf_with_input)
        exception = self.assertRaises(
            exc.InputException,
            t_s.create_cron_trigger,
            'trigger-%s' % utils.generate_unicode_uuid(),
            'some_wf',
            {},
            {},
            '*/5 * * * *',
            None,
            None,
            datetime.datetime(2010, 8, 25)
        )

        self.assertIn('Invalid input', str(exception))
        self.assertIn('some_wf', str(exception))

    def test_oneshot_trigger_create(self):
        trigger = t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            None,
            "4242-12-25 13:37",
            None,
            datetime.datetime(2010, 8, 25)
        )

        self.assertEqual(
            datetime.datetime(4242, 12, 25, 13, 37),
            trigger.next_execution_time
        )

    @mock.patch.object(security, 'create_trust',
                       type('trust', (object,), {'id': 'my_trust_id'}))
    def test_create_trust_in_trigger(self):
        cfg.CONF.set_default('auth_enable', True, group='pecan')
        self.addCleanup(
            cfg.CONF.set_default, 'auth_enable',
            False, group='pecan'
        )

        trigger = t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '*/2 * * * *',
            None,
            None,
            datetime.datetime(2010, 8, 25)
        )

        self.assertEqual('my_trust_id', trigger.trust_id)

    @mock.patch.object(security, 'create_trust',
                       type('trust', (object,), {'id': 'my_trust_id'}))
    @mock.patch.object(security, 'create_context')
    @mock.patch.object(rpc.EngineClient, 'start_workflow', mock.Mock())
    @mock.patch(
        'mistral.services.periodic.advance_cron_trigger',
        mock.MagicMock(side_effect=new_advance_cron_trigger)
    )
    @mock.patch.object(security, 'delete_trust')
    def test_create_delete_trust_in_trigger(self, delete_trust, create_ctx):
        create_ctx.return_value = self.ctx
        cfg.CONF.set_default('auth_enable', True, group='pecan')
        trigger_thread = periodic.setup()
        self.addCleanup(trigger_thread.stop)
        self.addCleanup(
            cfg.CONF.set_default, 'auth_enable',
            False, group='pecan'
        )

        t_s.create_cron_trigger(
            'trigger-%s' % utils.generate_unicode_uuid(),
            self.wf.name,
            {},
            {},
            '* * * * * *',
            None,
            1,
            datetime.datetime(2010, 8, 25)
        )

        eventlet.sleep(1)
        self.assertEqual(0, delete_trust.call_count)

    def test_get_trigger_in_correct_orders(self):
        t1_name = 'trigger-%s' % utils.generate_unicode_uuid()

        t_s.create_cron_trigger(
            t1_name,
            self.wf.name,
            {},
            pattern='*/5 * * * *',
            start_time=datetime.datetime(2010, 8, 25)
        )

        t2_name = 'trigger-%s' % utils.generate_unicode_uuid()

        t_s.create_cron_trigger(
            t2_name,
            self.wf.name,
            {},
            pattern='*/1 * * * *',
            start_time=datetime.datetime(2010, 8, 22)
        )

        t3_name = 'trigger-%s' % utils.generate_unicode_uuid()

        t_s.create_cron_trigger(
            t3_name,
            self.wf.name,
            {},
            pattern='*/2 * * * *',
            start_time=datetime.datetime(2010, 9, 21)
        )

        t4_name = 'trigger-%s' % utils.generate_unicode_uuid()

        t_s.create_cron_trigger(
            t4_name,
            self.wf.name,
            {},
            pattern='*/3 * * * *',
            start_time=datetime.datetime.utcnow() + datetime.timedelta(0, 50)
        )

        trigger_names = [t.name for t in t_s.get_next_cron_triggers()]

        self.assertEqual([t2_name, t1_name, t3_name], trigger_names)

    @mock.patch(
        'mistral.services.periodic.advance_cron_trigger',
        mock.MagicMock(side_effect=new_advance_cron_trigger)
    )
    @mock.patch.object(rpc.EngineClient, 'start_workflow')
    def test_single_execution_with_multiple_processes(self, start_wf_mock):
        def stop_thread_groups():
            print('Killing cron trigger threads...')
            [tg.stop() for tg in self.trigger_threads]

        self.trigger_threads = [
            periodic.setup(),
            periodic.setup(),
            periodic.setup()
        ]
        self.addCleanup(stop_thread_groups)

        trigger_count = 5
        t_s.create_cron_trigger(
            'ct1',
            self.wf.name,
            {},
            {},
            '* * * * * */1',  # Every second
            None,
            trigger_count,
            datetime.datetime(2010, 8, 25)
        )

        # Wait until there are 'trigger_count' executions.
        self._await(
            lambda: self._wait_for_single_execution_with_multiple_processes(
                trigger_count,
                start_wf_mock
            )
        )

        # Wait some more and make sure there are no more than 'trigger_count'
        # executions.
        eventlet.sleep(5)

        self.assertEqual(trigger_count, start_wf_mock.call_count)

    def _wait_for_single_execution_with_multiple_processes(self, trigger_count,
                                                           start_wf_mock):
        eventlet.sleep(1)

        return trigger_count == start_wf_mock.call_count

    def test_get_next_execution_time(self):
        pattern = '*/20 * * * *'
        start_time = datetime.datetime(2016, 3, 22, 23, 40)
        result = t_s.get_next_execution_time(pattern, start_time)
        self.assertEqual(result, datetime.datetime(2016, 3, 23, 0, 0))



# Copyright 2016 - Nokia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.services import workbooks as wb_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WORKBOOK_WITH_EXPRESSIONS = """
---
version: '2.0'

name: wb

workflows:
  test_tasks_function:
    input:
      - wf1_wx_id
      - wf2_wx_id
      - wf3_wx_id
      - wf4_wx_id
      - wf5_wx_id

    tasks:
      main_task:
        action: std.noop
        publish:
          all_tasks_yaql: <% tasks() %>
          all_tasks_jinja: "{{ tasks() }}"

          wf1_tasks_yaql: <% tasks($.wf1_wx_id) %>
          wf1_tasks_jinja: "{{ tasks(_.wf1_wx_id) }}"
          wf1_recursive_tasks_yaql: <% tasks($.wf1_wx_id, true) %>
          wf1_recursive_tasks_jinja: "{{ tasks(_.wf1_wx_id, true) }}"
          wf1_recursive_error_tasks_yaql: <% tasks($.wf1_wx_id, true, ERROR) %>
          wf1_recursive_error_tasks_jinja:
            "{{ tasks(_.wf1_wx_id, True, 'ERROR') }}"
          wf1_not_recursive_error_tasks_yaql:
            <% tasks($.wf1_wx_id, false, ERROR) %>
          wf1_not_recursive_error_tasks_jinja:
            "{{ tasks(_.wf1_wx_id, False, 'ERROR') }}"
          wf1_recursive_success_flat_tasks_yaql:
            <% tasks($.wf1_wx_id, true, SUCCESS, true) %>
          wf1_recursive_success_flat_tasks_jinja:
            "{{ tasks(_.wf1_wx_id, True, 'SUCCESS', True) }}"

          wf2_recursive_tasks_yaql: <% tasks($.wf2_wx_id, true) %>
          wf2_recursive_tasks_jinja: "{{ tasks(_.wf2_wx_id, true) }}"

          wf3_recursive_error_tasks_yaql: <% tasks($.wf3_wx_id, true, ERROR) %>
          wf3_recursive_error_tasks_jinja:
            "{{ tasks(_.wf3_wx_id, True, 'ERROR') }}"
          wf3_recursive_error_flat_tasks_yaql:
            <% tasks($.wf3_wx_id, true, ERROR, true) %>
          wf3_recursive_error_flat_tasks_jinja:
            "{{ tasks(_.wf3_wx_id, True, 'ERROR', True) }}"

          wf4_recursive_error_flat_tasks_yaql:
            <% tasks($.wf4_wx_id, true, ERROR, true) %>
          wf4_recursive_error_flat_tasks_jinja:
            "{{ tasks(_.wf4_wx_id, True, 'ERROR', True) }}"


          wf5_recursive_error_flat_tasks_yaql:
            <% tasks($.wf5_wx_id, true, ERROR, true) %>
          wf5_recursive_error_flat_tasks_jinja:
            "{{ tasks(_.wf5_wx_id, True, 'ERROR', True) }}"


  wf1_top_lvl:
    tasks:
      top_lvl_wf1_task_1:
        workflow: wf1_second_lvl

      top_lvl_wf1_task_2:
        action: std.noop

  wf1_second_lvl:
    tasks:
      second_lvl_wf1_task_1:
        workflow: wf1_third_lvl_fail
        on-error:
          - second_lvl_wf1_task_2

      second_lvl_wf1_task_2:
        action: std.noop

      second_lvl_wf1_task_3:
        action: std.noop

  wf1_third_lvl_fail:
    tasks:
      third_lvl_wf1_task_1:
        action: std.noop
        on-success:
          - third_lvl_wf1_task_2_fail

      third_lvl_wf1_task_2_fail:
        action: std.fail

      third_lvl_wf1_task_3:
        action: std.noop


  wf2_top_lvl:
    tasks:
      top_lvl_wf2_task_1:
        action: std.noop

      top_lvl_wf2_task_2:
        action: std.noop


  wf3_top_lvl:
    tasks:
      top_lvl_wf3_task_1_fail:
        workflow: wf3_second_lvl_fail

      top_lvl_wf3_task_2_fail:
        action: std.fail

  wf3_second_lvl_fail:
    tasks:
      second_lvl_wf3_task_1_fail:
        workflow: wf3_third_lvl_fail

      second_lvl_wf3_task_2:
        action: std.noop

      second_lvl_wf3_task_3:
        action: std.noop

  wf3_third_lvl_fail:
    tasks:
      third_lvl_wf3_task_1:
        action: std.noop
        on-success:
          - third_lvl_wf3_task_2

      third_lvl_wf3_task_2:
        action: std.noop

      third_lvl_wf3_task_3_fail:
        action: std.fail

  wf4_top_lvl:
    tasks:
      top_lvl_wf4_task_1:
        workflow: wf4_second_lvl
        publish:
          raise_error: <% $.invalid_yaql_expression %>

  wf4_second_lvl:
    tasks:
      second_lvl_wf4_task_1:
        action: std.noop

  wf5_top_lvl:
    tasks:
      top_lvl_wf5_task_1:
        workflow: wf4_second_lvl
        input:
          raise_error: <% $.invalid_yaql_expression2 %>

  wf5_second_lvl:
    tasks:
      second_lvl_wf5_task_1:
        workflow: wf5_third_lvl

  wf5_third_lvl:
    tasks:
      third_lvl_wf5_task_1:
        action: std.noop

"""


class TasksFunctionTest(base.EngineTestCase):
    def _assert_published_tasks(self, task, published_key,
                                expected_tasks_count=None,
                                expected_tasks_names=None):
        published = task.published[published_key]

        self.assertIsNotNone(
            published,
            "there is a problem with publishing '{}'".format(published_key)
        )

        published_names = [t['name'] for t in published]

        if expected_tasks_names:
            for e in expected_tasks_names:
                self.assertIn(e, published_names)

            if not expected_tasks_count:
                expected_tasks_count = len(expected_tasks_names)

        if expected_tasks_count:
            self.assertEqual(expected_tasks_count, len(published))

    def test_tasks_function(self):
        wb_service.create_workbook_v2(WORKBOOK_WITH_EXPRESSIONS)

        # Start helping workflow executions.
        wf1_ex = self.engine.start_workflow('wb.wf1_top_lvl')
        wf2_ex = self.engine.start_workflow('wb.wf2_top_lvl')
        wf3_ex = self.engine.start_workflow('wb.wf3_top_lvl')
        wf4_ex = self.engine.start_workflow('wb.wf4_top_lvl')
        wf5_ex = self.engine.start_workflow('wb.wf5_top_lvl')

        self.await_workflow_success(wf1_ex.id)
        self.await_workflow_success(wf2_ex.id)
        self.await_workflow_error(wf3_ex.id)
        self.await_workflow_error(wf4_ex.id)
        self.await_workflow_error(wf5_ex.id)

        # Start test workflow execution
        wf_ex = self.engine.start_workflow(
            'wb.test_tasks_function',
            wf_input={
                'wf1_wx_id': wf1_ex.id,
                'wf2_wx_id': wf2_ex.id,
                'wf3_wx_id': wf3_ex.id,
                'wf4_wx_id': wf4_ex.id,
                'wf5_wx_id': wf5_ex.id
            }
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)
        self.assertEqual(1, len(task_execs))

        main_task = task_execs[0]

        self._assert_published_tasks(main_task, 'all_tasks_yaql', 22)
        self._assert_published_tasks(main_task, 'all_tasks_jinja', 22)
        self._assert_published_tasks(
            main_task,
            'wf1_tasks_yaql',
            2,
            ['top_lvl_wf1_task_1', 'top_lvl_wf1_task_2']
        )

        self._assert_published_tasks(
            main_task,
            'wf1_tasks_jinja',
            2,
            ['top_lvl_wf1_task_1', 'top_lvl_wf1_task_2']
        )

        self._assert_published_tasks(
            main_task,
            'wf1_recursive_tasks_yaql',
            8,
            [
                'top_lvl_wf1_task_1',
                'top_lvl_wf1_task_2',
                'second_lvl_wf1_task_3',
                'second_lvl_wf1_task_1',
                'second_lvl_wf1_task_2',
                'third_lvl_wf1_task_3',
                'third_lvl_wf1_task_1',
                'third_lvl_wf1_task_2_fail'
            ]
        )

        self._assert_published_tasks(
            main_task,
            'wf1_recursive_tasks_jinja',
            8,
            [
                'top_lvl_wf1_task_1',
                'top_lvl_wf1_task_2',
                'second_lvl_wf1_task_3',
                'second_lvl_wf1_task_1',
                'second_lvl_wf1_task_2',
                'third_lvl_wf1_task_3',
                'third_lvl_wf1_task_1',
                'third_lvl_wf1_task_2_fail'
            ]
        )

        self._assert_published_tasks(
            main_task,
            'wf1_recursive_error_tasks_yaql',
            2,
            ['second_lvl_wf1_task_1', 'third_lvl_wf1_task_2_fail']
        )

        self._assert_published_tasks(
            main_task,
            'wf1_recursive_error_tasks_jinja',
            2,
            ['second_lvl_wf1_task_1', 'third_lvl_wf1_task_2_fail']
        )

        self._assert_published_tasks(
            main_task,
            'wf1_not_recursive_error_tasks_yaql',
            0
        )

        self._assert_published_tasks(
            main_task,
            'wf1_not_recursive_error_tasks_jinja',
            0
        )

        self._assert_published_tasks(
            main_task,
            'wf1_recursive_success_flat_tasks_yaql',
            5,
            [
                'top_lvl_wf1_task_2',
                'second_lvl_wf1_task_3',
                'second_lvl_wf1_task_2',
                'third_lvl_wf1_task_3',
                'third_lvl_wf1_task_1'
            ]
        )

        self._assert_published_tasks(
            main_task,
            'wf1_recursive_success_flat_tasks_jinja',
            5,
            [
                'top_lvl_wf1_task_2',
                'second_lvl_wf1_task_3',
                'second_lvl_wf1_task_2',
                'third_lvl_wf1_task_3',
                'third_lvl_wf1_task_1'
            ]
        )

        self._assert_published_tasks(
            main_task,
            'wf2_recursive_tasks_yaql',
            2,
            ['top_lvl_wf2_task_2', 'top_lvl_wf2_task_1']
        )

        self._assert_published_tasks(
            main_task,
            'wf2_recursive_tasks_jinja',
            2,
            ['top_lvl_wf2_task_2', 'top_lvl_wf2_task_1']
        )

        self._assert_published_tasks(
            main_task,
            'wf3_recursive_error_tasks_yaql',
            4,
            [
                'top_lvl_wf3_task_1_fail',
                'top_lvl_wf3_task_2_fail',
                'second_lvl_wf3_task_1_fail',
                'third_lvl_wf3_task_3_fail'
            ]
        )

        self._assert_published_tasks(
            main_task,
            'wf3_recursive_error_tasks_jinja',
            4,
            [
                'top_lvl_wf3_task_1_fail',
                'top_lvl_wf3_task_2_fail',
                'second_lvl_wf3_task_1_fail',
                'third_lvl_wf3_task_3_fail'
            ]
        )

        self._assert_published_tasks(
            main_task,
            'wf3_recursive_error_flat_tasks_yaql',
            2,
            ['top_lvl_wf3_task_2_fail', 'third_lvl_wf3_task_3_fail']
        )

        self._assert_published_tasks(
            main_task,
            'wf3_recursive_error_flat_tasks_jinja',
            2,
            ['top_lvl_wf3_task_2_fail', 'third_lvl_wf3_task_3_fail']
        )

        self._assert_published_tasks(
            main_task,
            'wf4_recursive_error_flat_tasks_yaql',
            1,
            ['top_lvl_wf4_task_1']
        )

        self._assert_published_tasks(
            main_task,
            'wf4_recursive_error_flat_tasks_jinja',
            1,
            ['top_lvl_wf4_task_1']
        )

        self._assert_published_tasks(
            main_task,
            'wf5_recursive_error_flat_tasks_yaql',
            1,
            ['top_lvl_wf5_task_1']
        )

        self._assert_published_tasks(
            main_task,
            'wf5_recursive_error_flat_tasks_jinja',
            1,
            ['top_lvl_wf5_task_1']
        )
